/*	$AoiMoe: $	*/

/*
 * Copyright (c) 2001 Takuya SHIOZAKI <AoiMoe@imou.to>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <inttypes.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/mman.h>

#include "peloader.h"


#define bool int
#define true 1
#define false 0

#define DPR(a) dprintf a

static void
dprintf(char const *format, ...)
{
	va_list ap;
	va_start(ap, format);
	vfprintf(stderr, format, ap);
	va_end(ap);
}


#define PE_BEGIN 0x3c
#define PE_SIGNATURE "PE\0\0"
#define PE_SIZE_OF_SIGNATURE 4
#define PE_SIZE_OF_COFF_FILE_HEADER 20
#define PE_SIZE_OF_OPTIONAL_HEADER 96
#define PE_SIZE_OF_IMAGE_DATA_DIRECTORY 8
#define PE_INDEX_OF_EXPORT_DIRECTORY 0

typedef struct {
	uint16_t machine;
	uint16_t numberOfSections;
	uint32_t timeDateStamp;
	uint32_t pointerToSymbolTable;
	uint32_t numberOfSymbols;
	uint32_t sizeOfOptionalHeader;
	uint32_t characteristic;
} PECoffFileHeader;


typedef struct {
	uint32_t rva;
	uint32_t size;
} PEImageDataDirectory;


typedef struct {
	// standard field
	uint16_t magic;
	uint8_t majorLinkerVersion;
	uint8_t minorLinkerVersion;
	uint32_t sizeOfCode;
	uint32_t sizeOfInitializedData;
	uint32_t sizeOfUninitializedData;
	uint32_t addressOfEntryPoint;
	uint32_t baseOfCode;
	uint32_t baseOfData;
	// NT specific field
	uint32_t imageBase;
	uint32_t sectionAlignment;
	uint32_t fileAlignment;
	uint16_t majorOperatingSystemVersion;
	uint16_t minorOperatingSystemVersion;
	uint16_t majorImageVersion;
	uint16_t minorImageVersion;
	uint16_t majorSubsystemVersion;
	uint16_t minorSubsystemVersion;
	//uint32_t reserved;
	uint32_t sizeOfImage;
	uint32_t sizeOfHeaders;
	uint32_t checkSum;
	uint16_t subsystem;
	uint16_t dllCharacteristics;
	uint32_t sizeOfStackReserve;
	uint32_t sizeOfStackCommit;
	uint32_t sizeOfHeapReserve;
	uint32_t sizeOfHeapCommit;
	uint32_t loaderFlags;
	uint32_t numberOfRvaAndSizes;
	PEImageDataDirectory *imageDirectory;
} PEOptionalHeader;

typedef struct {
	uint32_t exportFlags;
	uint32_t timeDateStamp;
	uint16_t majorVersion;
	uint16_t minorVersion;
	uint32_t nameRVA;
	uint32_t ordinalBase;
	uint32_t addressTableEntries;
	uint32_t numberOfNamePointers;
	uint32_t exportAddressTableRVA;
	uint32_t namePointerRVA;
	uint32_t ordinalTableRVA;
} PEExportDirectoryTable;

struct PEFile_ {
	uint8_t *mappedFile;
	size_t size;
	PECoffFileHeader *peCoffFileHeader;
	PEOptionalHeader *peOptionalHeader;
	PEExportDirectoryTable *peExportDirectoryTable;
};


static inline uint8_t mem_get8(uint8_t *p)
{
	return (*p);
}

static inline uint16_t mem_get16le(uint8_t *p)
{
	return (*p | ((uint16_t)*(p+1)<<8));
}

static inline uint32_t mem_get32le(uint8_t *p)
{
	return (*p |
		((uint32_t)*(p+1)<<8) |
		((uint32_t)*(p+2)<<16) |
		((uint32_t)*(p+3)<<24));
}

#define MEM_GET8(dst_, p_) \
	do { (dst_) = mem_get8(p); p++; } while (/*CONSTCOND*/0)
#define MEM_GET16LE(dst_, p_) \
	do { (dst_) = mem_get16le(p); p+=2; } while (/*CONSTCOND*/0)
#define MEM_GET32LE(dst_, p_) \
	do { (dst_) = mem_get32le(p); p+=4; } while (/*CONSTCOND*/0)

static void get_coff_file_header(PECoffFileHeader *pch, uint8_t *p)
{
	pch->machine = mem_get16le(p+0);
	pch->numberOfSections = mem_get16le(p+2);
	pch->timeDateStamp = mem_get32le(p+4);
	pch->pointerToSymbolTable = mem_get32le(p+8);
	pch->numberOfSymbols = mem_get32le(p+12);
	pch->sizeOfOptionalHeader = mem_get32le(p+16);
	pch->characteristic = mem_get32le(p+18);
}

static void get_optional_header(PEOptionalHeader *poh, uint8_t *p)
{
	poh->magic = mem_get16le(p+0);
	poh->majorLinkerVersion = mem_get8(p+2);
	poh->minorLinkerVersion = mem_get8(p+3);
	poh->sizeOfCode = mem_get32le(p+4);
	poh->sizeOfInitializedData = mem_get32le(p+8);
	poh->sizeOfUninitializedData = mem_get32le(p+12);
	poh->addressOfEntryPoint = mem_get32le(p+16);
	poh->baseOfCode = mem_get32le(p+20);
	poh->baseOfData = mem_get32le(p+24);
	// NT Specific
	poh->imageBase = mem_get32le(p+28);
	poh->sectionAlignment = mem_get32le(p+32);
	poh->fileAlignment = mem_get32le(p+36);
	poh->majorOperatingSystemVersion = mem_get16le(p+40);
	poh->minorOperatingSystemVersion = mem_get16le(p+42);
	poh->majorImageVersion = mem_get16le(p+44);
	poh->minorImageVersion = mem_get16le(p+46);
	poh->majorSubsystemVersion = mem_get16le(p+48);
	poh->minorSubsystemVersion = mem_get16le(p+50);
	//poh->reserved = mem_get32le(p+52);
	poh->sizeOfImage = mem_get32le(p+56);
	poh->sizeOfHeaders = mem_get32le(p+60);
	poh->checkSum = mem_get32le(p+64);
	poh->subsystem = mem_get16le(p+68);
	poh->dllCharacteristics = mem_get16le(p+70);
	poh->sizeOfStackReserve = mem_get32le(p+72);
	poh->sizeOfStackCommit = mem_get32le(p+76);
	poh->sizeOfHeapReserve = mem_get32le(p+80);
	poh->sizeOfHeapCommit = mem_get32le(p+84);
	poh->loaderFlags = mem_get32le(p+88);
	poh->numberOfRvaAndSizes = mem_get32le(p+92);
}


static void get_image_data_directory(PEImageDataDirectory *img, uint8_t *p)
{
	img->rva  = mem_get32le(p+0);
	img->size = mem_get32le(p+4);
}

static void get_export_directory_table(PEExportDirectoryTable *edt, uint8_t *p)
{
	edt->exportFlags = mem_get32le(p+0);
	edt->timeDateStamp = mem_get32le(p+4);
	edt->majorVersion = mem_get16le(p+8);
	edt->minorVersion = mem_get16le(p+10);
	edt->nameRVA = mem_get32le(p+12);
	edt->ordinalBase = mem_get32le(p+16);
	edt->addressTableEntries = mem_get32le(p+20);
	edt->numberOfNamePointers = mem_get32le(p+24);
	edt->exportAddressTableRVA = mem_get32le(p+28);
	edt->namePointerRVA = mem_get32le(p+32);
	edt->ordinalTableRVA = mem_get32le(p+36);
}


void *find_export_symbol_byname(PEFile *pe, char const *name)
{
	uint32_t i, num, rva;
	uint32_t obase, ord;
	uint8_t *eat, *npt, *odt, *n;

	if (!pe->peExportDirectoryTable)
		return;

	num = pe->peExportDirectoryTable->numberOfNamePointers;
	eat = pe->mappedFile+pe->peExportDirectoryTable->exportAddressTableRVA;
	npt = pe->mappedFile+pe->peExportDirectoryTable->namePointerRVA;
	odt = pe->mappedFile+pe->peExportDirectoryTable->ordinalTableRVA;
	obase = pe->peExportDirectoryTable->ordinalBase;
	for (i=0; i<num; i++) {
		n = pe->mappedFile + mem_get32le(npt);
		if (!strcmp(n, name)) {
			// match
			ord = mem_get16le(odt);
			rva = mem_get32le(eat+ord*4);
			DPR(("%s: ord=%u, rva=0x%08X\n",
			     name, ord+obase, rva));
			return (void *)(pe->mappedFile+rva);
		}
		npt += 4;
		odt += 2;
	}
	return 0;
}


PEFile *
map_pecoff(char const *file)
{
	PEFile *pe;
	int fd;
	void *map;
	struct stat sb;
	uint8_t *p;
	uint32_t num, i, pebegin;
	PEImageDataDirectory *img;

	errno = 0;
	fd = open(file, O_RDONLY);
	if (fd<0) {
		DPR(("cannot open file %s : %s\n", file, strerror(errno)));
		return NULL;
	}
	if (fstat(fd, &sb)) {
		DPR(("cannot get stat of file %s : %s\n",
		     file, strerror(errno)));
		close(fd);
		return NULL;
	}
	map = mmap(0, sb.st_size, PROT_READ, MAP_FILE | MAP_PRIVATE, fd, 0);
	if (!map) {
		DPR(("cannot map file %s : %s\n", file, strerror(errno)));
		close(fd);
		return NULL;
	}
	close(fd);
	if (sb.st_size<PE_BEGIN+4 ||
	    sb.st_size<((pebegin=mem_get32le(map+PE_BEGIN))+
			PE_SIZE_OF_SIGNATURE)) {
		DPR(("file %s is too small\n", file));
		munmap(map, sb.st_size);
		return NULL;
	}
	;
	if (memcmp((char const *)map+pebegin,
		   PE_SIGNATURE, PE_SIZE_OF_SIGNATURE)) {
		DPR(("unmatch signature\n"));
		munmap(map, sb.st_size);
		return NULL;
	}
	pe = calloc(1, sizeof(PEFile));
	pe->mappedFile = map;
	pe->size = sb.st_size;
	pe->peCoffFileHeader = calloc(1, sizeof(PECoffFileHeader));
	pe->peOptionalHeader = calloc(1, sizeof(PEOptionalHeader));
	pe->peExportDirectoryTable = calloc(1, sizeof(PEExportDirectoryTable));

	p = map+pebegin+4;
	get_coff_file_header(pe->peCoffFileHeader, p);
	p += PE_SIZE_OF_COFF_FILE_HEADER;
	get_optional_header(pe->peOptionalHeader, p);
	p += PE_SIZE_OF_OPTIONAL_HEADER;

	num = pe->peOptionalHeader->numberOfRvaAndSizes;
	img = pe->peOptionalHeader->imageDirectory =
	    calloc(num, sizeof(PEImageDataDirectory));
	for (i=0; i<pe->peOptionalHeader->numberOfRvaAndSizes; i++) {
		get_image_data_directory(img, p);
		img++; p+=PE_SIZE_OF_IMAGE_DATA_DIRECTORY;
	}

	get_export_directory_table(
		pe->peExportDirectoryTable,
		map + pe->peOptionalHeader->imageDirectory[
			PE_INDEX_OF_EXPORT_DIRECTORY].rva);

	return pe;
}

void
unmap_pecoff(PEFile *pe)
{
	free(pe->peOptionalHeader->imageDirectory);
	free(pe->peExportDirectoryTable);
	free(pe->peOptionalHeader);
	free(pe->peCoffFileHeader);
	munmap(pe->mappedFile, pe->size);
	free(pe);
}


#ifdef TEST
int
main()
{
	PEFile *pe;
	uint8_t *p;
	int i, j, k;

	pe = map_pecoff("UnivUI.dll");
	p = find_export_symbol_byname(
		pe, "?_GLOBAL_tblQuantTransfer@@3PAUCQuantTbl@@A");
	DPR(("p=%p\n", p));
	for (i=0; i<6; i++) {
		printf("{\n");
		for (j=0; j<256; j+=8) {
			printf("\t");
			for (k=0; k<8; k++) {
				printf("0x%02X, ", *p);
				p++;
			}
			printf("\n");
		}
		printf("},\n");
	}
	unmap_pecoff(pe);
	
	return 0;
}
#endif
